import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestQuant(unittest.TestCase):
    """
    random test on the behavior of nearest fixed point rounding
    """
    def test_fixed_random(self):
        S = lambda bits :  2**(bits)
        Q = lambda x, bits : torch.round(x*S(bits))/S(bits)
        wl = 8
        quant = lambda x : fixed_point_quantize(x, wl=wl, fl=wl, clamp=False, rounding="nearest")

        N = int(1e8)
        x = torch.randn(N, device='cuda')
        oracle = Q(x, wl)
        target = quant(x)
        matched = torch.eq(oracle, target).all().item()
        self.assertTrue(matched)

if __name__ == "__main__":
    unittest.main()



__all__ = ['Number', 'FixedPoint', 'BlockFloatingPoint', 'FloatingPoint']

class Number():
    def __init__(self):
        pass

    def __str__(self):
        raise NotImplemented

class FixedPoint(Number):
    r"""
    Low-Precision Fixed Point Format. Defined similarly in
    *Deep Learning with Limited Numerical Precision* (https://arxiv.org/abs/1502.02551)

    The representable range is :math:`[-2^{wl-fl-1}, 2^{wl-fl-1}-2^{-fl}]`
    and a precision unit (smallest nonzero absolute value) is
    :math:`2^{-fl}`.
    Numbers outside of the representable range can be clamped
    (if `clamp` is true).
    We can also give up the smallest representable number to make the range
    symmetric, :math:`[-2^{wl-fl-1}^{-fl}, 2^{wl-fl-1}-2^{-fl}]`. (if `symmetric` is true).

    Define :math:`\lfloor x \rfloor` to be the largest representable number (multiples of :math:`2^{-fl}`) smaller than :math:`x`.
    For numbers within the representable range, fixed point quantizatio corresponds to

    .. math::

       NearestRound(x)
       =
       \Biggl \lbrace
       {
       \lfloor x \rfloor, \text{ if } \lfloor x \rfloor \leq x \leq \lfloor x \rfloor + 2^{-fl-1}
       \atop
        \lfloor x \rfloor + 2^{-fl}, \text{ if } \lfloor x \rfloor + 2^{-fl-1} < x \leq \lfloor x \rfloor + 2^{-fl}
       }

    or

    .. math::
       StochasticRound(x)
       =
       \Biggl \lbrace
       {
       \lfloor x \rfloor, \text{ with probabilty } 1 - \frac{x - \lfloor x \rfloor}{2^{-fl}}
       \atop
        \lfloor x \rfloor + 2^{-fl}, \text{ with probabilty } \frac{x - \lfloor x \rfloor}{2^{-fl}}
       }

    Args:
        - :attr: wl (int) : word length of each fixed point number
        - :attr: fl (int) : fractional length of each fixed point number
        - :attr: clamp (bool) : whether to clamp unrepresentable numbers
        - :attr: symmetric (bool) : whether to make the representable range symmetric
    """

    def __init__(self, wl, fl, clamp=True, symmetric=False):
        assert wl > 0, "invalid bits for word length: {}".format(wl)
        assert fl > 0, "invalid bits for fractional length: {}".format(fl)
        assert type(symmetric) == bool, "invalid type for clamping choice: {}".format(type(clamp))
        assert type(symmetric) == bool, "invalid type for symmetric: {}".format(type(symmetric))
        self.wl = wl
        self.fl = fl
        self.clamp = clamp
        self.symmetric = symmetric

    def __str__(self):
        return "FixedPoint (wl={:d}, fl={:d})".format(self.wl, self.fl)

    def __repr__(self):
        return "FixedPoint (wl={:d}, fl={:d})".format(self.wl, self.fl)

class FloatingPoint(Number):
    """
    Low-Precision Floating Point Format.

    We set the exponent bias to be :math:`2^{exp-1}`. In our simulation, we do
    not handle denormal/subnormal numbers and infinities/NaNs. For rounding
    mode, we apply *round to nearest even*.

    Args:
        - :attr: `exp`: number of bits allocated for exponent
        - :attr: `man`: number of bits allocated for mantissa, referring to number of bits that are
                        supposed to be stored on hardware (not counting the virtual bits).
    """
    def __init__(self, exp, man):
        assert 8 >= exp > 0, "invalid bits for exponent:{}".format(exp)
        assert 23 >= man > 0, "invalid bits for mantissa:{}".format(man)
        self.exp = exp
        self.man = man

    def __str__(self):
        return "FloatingPoint (exponent={:d}, mantissa={:d})".format(self.exp, self.man)

    def __repr__(self):
        return "FloatingPoint (exponent={:d}, mantissa={:d})".format(self.exp, self.man)

class BlockFloatingPoint(Number):
    """
    Low-Precision Block Floating Point Format.

    BlockFloatingPoint shares an exponent across a block of numbers. The shared exponent is chosen from
    the largest magnitude in the block.

    Args:
        - :attr: `wl` word length of the tensor
        - :attr: `dim` block dimension to share exponent. (*, D, *) Tensor where
          D is at position `dim` will have D different exponents; use -1 if the
          entire tensor is treated as a single block (there is only 1 shared
          exponent).
    """
    def __init__(self, wl, dim=-1):
        assert wl > 0 and isinstance(wl, int), "invalid bits for word length:{}".format(wl)
        assert dim >= -1 and isinstance(dim, int), "invalid dimension"
        self.wl = wl
        self.dim = dim

    def __str__(self):
        return "BlockFloatingPoint (wl={:d}, dim={:d})".format(self.wl, self.dim)

    def __repr__(self):
        return "BlockFloatingPoint (wl={:d}, dim={:d})".format(self.wl, self.dim)



import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestBackward(unittest.TestCase):
    """
    invariant: stochastic rounding is unbiased
    """
    def test_backward_fixed(self):
        for d in ['cpu', 'cuda']:
            number = FixedPoint(wl=5, fl=4)
            t_min = -2 ** (number.wl-number.fl-1)
            t_max = 2 ** (number.wl-number.fl-1) - 2 ** (-number.fl)
            a = torch.linspace(t_min, t_max, steps=100, device=d, requires_grad=True)
            quant = quantizer(forward_number=number, forward_rounding='nearest', backward_number=number, backward_rounding='nearest')
            s = quant(a).sum()
            s.backward()
            true_grad = torch.ones_like(a)*t_max
            self.assertTrue(torch.eq(a.grad, true_grad).all().item())

            a = torch.tensor([-1, -0.6, 0, 1], device=d, requires_grad=True)
            quant = quantizer(forward_number=number, forward_rounding='nearest', backward_number=number, backward_rounding='nearest', clamping_grad_zero=True)
            s = quant(a).sum()
            s.backward()
            true_grad = torch.ones_like(a)*t_max
            true_grad[-1] = 0
            self.assertTrue(torch.eq(a.grad, true_grad).all().item())

    def test_clamp_zero(self):
        def S(bits):
            return 2.**(bits-1)

        def C(x, bits):
            if bits > 15 or bits == 1:
                delta = 0
            else:
                delta = 1. / S(bits)
                upper = 1  - delta
                lower = -1 + delta
                return torch.clamp(x, lower, upper)

        def Q(x, bits):
            assert bits != -1
            if bits==1:
                return torch.sign(x)
            if bits > 15:
                return x
            return torch.round(x*S(bits))/S(bits)

        class Rounding(torch.autograd.Function):
            @staticmethod
            def forward(self, x, bits_A, bits_E):
                self.bits_E = bits_E
                x = Q(x, bits_A)
                t_max = 1- 1./(2.**(bits_A-1))
                t_min = -1 + 1./(2.**(bits_A-1))
                mask = (x > t_max) + (x < t_min)
                x = torch.clamp(x, t_min, t_max)
                self.mask = mask

                return x

            @staticmethod
            def backward(self, grad_output):
                if self.needs_input_grad[0]:
                    grad_input = Q(C(grad_output, self.bits_E), self.bits_E).masked_fill_(self.mask, 0)
                return grad_input, None, None

        wl = 4
        number = FixedPoint(wl=wl, fl=wl-1, clamp=True, symmetric=True)
        oracle = lambda x : Rounding.apply(x, wl, wl)
        quant = quantizer(forward_number=number, forward_rounding='nearest',
                          backward_number=number, backward_rounding='nearest',
                          clamping_grad_zero=True)

        x = torch.linspace(-2, 2, steps=10, device='cpu', requires_grad=True)
        x.sum().backward()

        y = torch.linspace(-2, 2, steps=10, device='cpu', requires_grad=True)
        oracle(y).sum().backward()

        z = torch.linspace(-2, 2, steps=10, device='cpu', requires_grad=True)
        quant(z).sum().backward()

        self.assertTrue(torch.eq(y.grad, z.grad).all().item())

    def test_backward_block(self):
        for d in ['cpu', 'cuda']:
            number = BlockFloatingPoint(wl=5)
            a = torch.linspace(-0.9, 0.9, steps=100, device=d, requires_grad=True)
            quant = quantizer(forward_number=number, forward_rounding='nearest', backward_number=number, backward_rounding="nearest")
            s = quant(a).sum()
            s.backward()
            true_grad = torch.ones_like(a)
            self.assertTrue(torch.eq(a.grad, true_grad).all().item())

    def test_backward_float(self):
        for d in ['cpu', 'cuda']:
            number = FloatingPoint(exp=3, man=5)
            a = torch.linspace(-0.9, 0.9, steps=100, device=d, requires_grad=True)
            quant = quantizer(forward_number=number, forward_rounding='nearest', backward_number=number, backward_rounding="nearest")
            s = quant(a).sum()
            s.backward()
            true_grad = torch.ones_like(a)
            self.assertTrue(torch.eq(a.grad, true_grad).all().item())

if __name__ == "__main__":
    unittest.main()



import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestStochastic(unittest.TestCase):
    """
    invariant: quantized numbers cannot be greater than the maximum representable number
    or lower than the maximum representable number
    """
    def test_fixed(self):
        """test fixed point clamping"""
        for d in ['cpu', 'cuda']:
            for r in ['stochastic', "nearest"]:
                wl = 5
                fl = 4
                t_min = - 2 ** (wl-fl-1)
                t_max = 2 ** (wl-fl-1) - 2 ** (-fl)
                a = torch.linspace(-2, 2, steps=100, device=d)
                clamp_a = fixed_point_quantize(a, wl=wl, fl=fl, clamp=True, rounding=r)
                self.assertEqual(t_max, clamp_a.max().item())
                self.assertEqual(t_min, clamp_a.min().item())

                a = torch.linspace(-2, 2, steps=100, device=d)
                no_clamp_a = fixed_point_quantize(a, wl=wl, fl=fl, clamp=False, rounding=r)
                self.assertLess(t_max, no_clamp_a.max().item())
                self.assertGreater(t_min, no_clamp_a.min().item())

    def test_float(self):
        """test floating point clamping"""
        formats=[(6, 9), (5, 10), (5, 2)]

        for exp, man in formats:
            for d in ['cpu', 'cuda']:
                for r in ['stochastic', "nearest"]:
                    # test positive
                    a_max = 2**(2**(exp-1))*(1-2**(-man-1))
                    a_min = 2**(-2**(exp-1)+1)
                    a = torch.Tensor([2**50, 2**(-50)]).to(device=d)
                    quant_a = float_quantize(a, exp=exp, man=man, rounding=r)
                    self.assertEqual(quant_a[0].item(), a_max)
                    self.assertAlmostEqual(quant_a[1].item(), a_min)

                    # test negative
                    a_max = -2**(2**(exp-1))*(1-2**(-man-1))
                    a_min = -2**(-2**(exp-1)+1)
                    a = torch.Tensor([-2**35, -2**(-35)]).to(device=d)
                    quant_a = float_quantize(a, exp=exp, man=man, rounding=r)
                    self.assertEqual(quant_a[0].item(), a_max)
                    self.assertAlmostEqual(quant_a[1].item(), a_min)

if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
from qtorch.quant import *
from collections import OrderedDict
import copy

__all__ =  ['lower', 'sequential_lower']

SEQUENTIAL_LAYERS = [nn.Sequential, nn.ModuleList] # TODO: Param List

DICT_LAYERS = [nn.ModuleDict]

CONV_LAYERS = [nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d,
               nn.ConvTranspose3d, nn.Unfold, nn.Fold]

POOL_LAYERS = [nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d, nn.MaxUnpool1d, nn.MaxUnpool2d,
               nn.MaxUnpool3d, nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d,
               nn.FractionalMaxPool2d, nn.LPPool1d, nn.LPPool2d, nn.AdaptiveMaxPool1d,
               nn.AdaptiveMaxPool2d, nn.AdaptiveAvgPool2d,
               nn.AdaptiveMaxPool1d, nn.AdaptiveAvgPool1d,
               nn.AdaptiveMaxPool3d, nn.AdaptiveAvgPool3d]

PAD_LAYERS = [nn.ReflectionPad1d, nn.ReflectionPad2d,
              nn.ReplicationPad1d, nn.ReplicationPad2d,
              nn.ZeroPad2d,
              nn.ConstantPad1d, nn.ConstantPad2d, nn.ConstantPad3d]

ACTIVATION_LAYERS = [nn.ELU, nn.Hardshrink, nn.Hardtanh, nn.LeakyReLU, nn.LogSigmoid,
                     nn.PReLU, nn.ReLU, nn.ReLU6, nn.RReLU,
                     nn.SELU, nn.Sigmoid, nn.Softplus, nn.Softshrink,
                     nn.Softsign, nn.Tanh, nn.Tanhshrink, nn.Threshold,
                     nn.Softmin, nn.Softmax, nn.Softmax2d, nn.LogSoftmax,
                    ]#nn.AdaptiveLogSoftmaxWithLoss]

NORM_LAYERS = [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
               nn.GroupNorm, nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d,
               nn.LayerNorm, nn.LocalResponseNorm]

# Not supporting RNN layer

LINEAR_LAYERS = [nn.Linear, nn.Bilinear]

DROPOUT_LAYERS = [nn.Dropout, nn.Dropout2d, nn.Dropout3d, nn.AlphaDropout]

# Not supporting Sparse/Distance layers

LOSS_LAYERS = [nn.L1Loss, nn.MSELoss, nn.CrossEntropyLoss, nn.NLLLoss, nn.PoissonNLLLoss,
               nn.KLDivLoss, nn.BCELoss, nn.BCEWithLogitsLoss, nn.MarginRankingLoss,
               nn.HingeEmbeddingLoss, nn.MultiLabelMarginLoss, nn.SmoothL1Loss,
               nn.SoftMarginLoss, nn.MultiLabelSoftMarginLoss, #nn.CosineEmbeddingLos,
               nn.MultiMarginLoss, nn.TripletMarginLoss]

LAYERS_TYPES = {
                    "conv":CONV_LAYERS,
                    "linear":LINEAR_LAYERS,
                    "pool":POOL_LAYERS,
                    "pad":PAD_LAYERS,
                    "activation":ACTIVATION_LAYERS,
                    "normalization":NORM_LAYERS,
                    "dropout":DROPOUT_LAYERS,
                    "loss":LOSS_LAYERS
               }

def _get_apply_lower_func(quant, layer_types=[]):
    def _insert_LP_layer(module):
        """Insert quant layer for all layers so long as in layer_types
        """
        lp_layer_types = []
        for layer_type in layer_types:
            assert layer_type in LAYERS_TYPES.keys()
            lp_layer_types += LAYERS_TYPES[layer_type]

        old_forward = module.forward
        if type(module) in lp_layer_types:
            module.forward = lambda *input : quant(old_forward(*input))
        else:
            return
    return _insert_LP_layer

def _get_return_sequential_lower_func(quant, layer_types=[]):
    def _insert_LP_layer(module):
        """Insert quant layer for all layers so long as in layer_types
        """
        if type(module) in SEQUENTIAL_LAYERS:
            for i, sub_module in enumerate(module.children()):
                module[i] = _insert_LP_layer(module[i])
            return module
        elif type(module) in DICT_LAYERS:
            for key, sub_module in module.items():
                module[key] = _insert_LP_layer(module[key])
            return module
        elif len(list(module.children())) != 0:
            for attribute_name in module.__dir__():
                attribute = getattr(module, attribute_name)
                if isinstance(attribute, nn.Module):
                    setattr(module, attribute_name, _insert_LP_layer(attribute))
            return module
        else:
            # Module has no submodule
            lp_layer_types = []
            for layer_type in layer_types:
                assert layer_type in LAYERS_TYPES.keys()
                lp_layer_types += LAYERS_TYPES[layer_type]

            if type(module) in lp_layer_types:
                module = nn.Sequential(module, quant)
            return module

    return _insert_LP_layer
 

def lower(model,
          layer_types=[],
          forward_number=None, backward_number=None,
          forward_rounding="stochastic", backward_rounding="stochastic"):
    quant = Quantizer(forward_number, backward_number,
                      forward_rounding, backward_rounding)
    lower_func = _get_apply_lower_func(quant, layer_types=layer_types)
    model.apply(lower_func)


def sequential_lower(model,
                     layer_types=[],
                     forward_number=None, backward_number=None,
                     forward_rounding="stochastic", backward_rounding="stochastic"):
    """Return a new model without touching the old one
    """
    quant = Quantizer(forward_number, backward_number,
                      forward_rounding, backward_rounding)
    
    lower_func = _get_return_sequential_lower_func(quant, layer_types=layer_types)
    return lower_func(copy.deepcopy(model))

# def sequential_lower_(model,
#                       layer_types=[],
#                       forward_number=None, backward_number=None,
#                       forward_rounding="stochastic", backward_rounding="stochastic"):
#     """Modify the original model
#     """
#     quant = Quantizer(forward_number, backward_number,
#                       forward_rounding, backward_rounding)
    
#     lower_func = _get_return_sequential_lower_func(quant, layer_types=layer_types)
#     return _get_return_sequential_lower_func(quant, layer_types=layer_types)(model)






import torch
import unittest
from qtorch.quant import block_quantize, fixed_point_quantize, float_quantize
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestRelation(unittest.TestCase):
    def test_fixed_block_zero_exponent(self):
        """
        invariant: when the max exponent of a block is zero, block floating point behaves similar to
        a fixed point where fl = wl -1, without the lowest number of the fixed point (-1).
        """
        for wl, fl in [(3,2), (5,4)]:
            t_max = 1-(2**(-fl))
            to_quantize = torch.linspace(-t_max, t_max, steps=1000, device='cuda')
            fixed_quantized = fixed_point_quantize(to_quantize, wl=wl, fl=fl, rounding='nearest')
            block_quantized = block_quantize(to_quantize, wl=wl, rounding='nearest')
            self.assertTrue(torch.eq(fixed_quantized, block_quantized).all().item())

    # def test_block_float_same_exponent(self):
    #     """
    #     invariant: when there is only one kind of exponent in a block, block floating point behaves the same as
    #     a floating point
    #     """
    #     self.assertTrue(False), "bad"

    def test_float_half(self):
        """
        invariant: when quantizing into a half precision floating point, float quantize kernel behaves the same as the
        built-in half precision tensor
        """
        a = torch.rand(int(1e4)).cuda()
        man = 10
        exp = 5
        half_a = a.half()
        def compute_dist(a, half_a):
            diff =  (a-half_a.float())
            diff2 = diff**2
            return diff2.sum()
        sim_half_a = float_quantize(a, exp=exp, man=man, rounding="nearest")
        self.assertEqual(compute_dist(a, sim_half_a), compute_dist(a, half_a))

if __name__ == "__main__":
    unittest.main()



import torch
import unittest
import qtorch
from qtorch.quant import block_quantize

def block_share_first(data, bits, ebit=8):
    """
    Sample implementation of sharing first dimension nearest rounding block floating point.
    """
    max_entry = torch.max(torch.abs(data.view(data.size(0), -1)), 1)[0]
    max_exponent = torch.floor(torch.log2(max_entry))
    # max_exponent = torch.clamp(max_exponent, -2**(ebit-1), 2**(ebit-1)-1)
    max_exponent = max_exponent.view([data.size(0)]+[1 for _ in range(data.dim()-1)])
    i = data * 2**(-max_exponent+(bits-2))
    i.round_()
    i.clamp_(-2**(bits-1), 2**(bits-1)-1)
    temp = i * 2**(max_exponent-(bits-2))
    return temp

class TestPython(unittest.TestCase):
    """
    QPyTorch Implementation should perform the same with PyTorch implementation
    """
    def test_block_sharing(self):
        pass
        # target = torch.ones(1)*0.99
        # for wl in [5]:
        #     python_result = block_share_first(target, wl)
        #     qtorch_result = block_quantize(target, wl=wl, dim=0, rounding="nearest")
        #     eq_mask = python_result.eq(qtorch_result)
        #     # print(python_result.masked_select(1-eq_mask))
        #     # print(qtorch_result.masked_select(1-eq_mask))
        #     mse = torch.sum(diff**2)
        #     # print(mse)

if __name__ == "__main__":
    unittest.main()



import torch
import unittest
import qtorch
from qtorch.quant import block_quantize, fixed_point_quantize, float_quantize
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestDevice(unittest.TestCase):
    """
        invariant: cuda and cpp implementation should behave the same
    """
    def error(self, cuda_t, cpu_t):
        return ((cuda_t.cpu()-cpu_t)**2).sum().item()

    def test_fixed_point(self):
        for wl, fl in [(5,4), (3,2)]:
            for rounding in ["nearest"]:
                t_max = 1-(2**(-fl))
                to_quantize_cuda = torch.linspace(-t_max, t_max, steps=20, device='cuda')
                to_quantize_cpu = to_quantize_cuda.clone().to("cpu")
                fixed_quantized_cuda = fixed_point_quantize(to_quantize_cuda, wl=wl, fl=fl, rounding=rounding)
                fixed_quantized_cpu = fixed_point_quantize(to_quantize_cpu, wl=wl, fl=fl, rounding=rounding)
                mse = self.error(fixed_quantized_cuda, fixed_quantized_cpu)
                self.assertTrue(mse<1e-15)
                # self.assertTrue(torch.eq(fixed_quantized_cuda.cpu(), fixed_quantized_cpu).all().item())

    def test_block_floating_point(self):
        for wl in [5, 3]:
            for rounding in ["nearest"]:
                for dim in [-1, 0, 1]:
                    t_max = 1-(2**(-4))
                    to_quantize_cuda = torch.linspace(-t_max, t_max, steps=20, device='cuda')
                    to_quantize_cpu = to_quantize_cuda.clone().to("cpu")
                    block_quantized_cuda = block_quantize(to_quantize_cuda, wl=wl, rounding=rounding)
                    block_quantized_cpu =  block_quantize(to_quantize_cpu,  wl=wl, rounding=rounding)
                    mse = self.error(block_quantized_cuda, block_quantized_cpu)
                    self.assertTrue(mse<1e-15)
                    # self.assertTrue(torch.eq(block_quantized_cuda.cpu(), block_quantized_cpu).all().item())

    def test_floating_point(self):
        for man, exp in [(2, 5), (6, 9)]:
            for rounding in ["nearest"]:
                to_quantize_cuda = torch.rand(20).cuda()
                to_quantize_cpu = to_quantize_cuda.clone().to("cpu")
                float_quantized_cuda = float_quantize(to_quantize_cuda, man=man, exp=exp, rounding=rounding)
                float_quantized_cpu  = float_quantize(to_quantize_cpu,  man=man, exp=exp, rounding=rounding)
                mse = self.error(float_quantized_cuda, float_quantized_cpu)
                self.assertTrue(mse<1e-15)

if __name__ == "__main__":
    unittest.main()



import torch
import torch.nn as nn
from .quant_function import *
import numpy as np

__all__ = ['Quantizer']

class Quantizer(nn.Module):
    def __init__(self, forward_number=None, backward_number=None,
                 forward_rounding="stochastic", backward_rounding="stochastic"):
        super(Quantizer, self).__init__()
        self.quantize = quantizer(forward_number, backward_number,
                                 forward_rounding, backward_rounding)

    def forward(self, x):
        return self.quantize(x)



from setuptools import setup, find_packages

try:
    import torch
    has_dev_pytorch = "dev" in torch.__version__
except ImportError:
    has_dev_pytorch = False

# Base equirements
install_requires = [
    "torch>=1.0.0",
]
if has_dev_pytorch:  # Remove the PyTorch requirement
    install_requires = [
        install_require for install_require in install_requires
        if "torch" != re.split(r"(=|<|>)", install_require)[0]
    ]

setup(name='qtorch',
      version='0.1.1',
      description="Low-Precision Arithmetic Simulation in Pytorch",
      long_description=open("README.md").read(),
      author="Tianyi Zhang, Zhiqiu Lin, Guandao Yang, Christopher De Sa",
      author_email="tz58@cornell.edu",
      project_urls={
        "Documentation": "https://qpytorch.readthedocs.io",
        "Source": "https://github.com/Tiiiger/QPyTorch/graphs/contributors",
      },
      packages=find_packages(),
      include_package_data=True,
      license="MIT",
      python_requires=">=3.6",
      install_requires=install_requires
)



import torch
from qtorch import Number, FixedPoint, BlockFloatingPoint, FloatingPoint
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.utils.cpp_extension import load
import os
current_path = os.path.dirname(os.path.realpath(__file__))
quant_cpu = load(
        name='quant_cpu',
        sources=[
            os.path.join(current_path, "quant_cpu/quant_cpu.cpp"),
            os.path.join(current_path, "quant_cpu/bit_helper.cpp"),
            os.path.join(current_path, "quant_cpu/sim_helper.cpp"),
            ]
        )

if torch.cuda.is_available():
     quant_cuda = load(
             name='quant_cuda',
             sources=[
                 os.path.join(current_path, "quant_cuda/quant_cuda.cpp"),
                 os.path.join(current_path, "quant_cuda/bit_helper.cu"),
                 os.path.join(current_path, "quant_cuda/sim_helper.cu"),
                 os.path.join(current_path, "quant_cuda/block_kernel.cu"),
                 os.path.join(current_path, "quant_cuda/float_kernel.cu"),
                 os.path.join(current_path, "quant_cuda/fixed_point_kernel.cu"),
                 os.path.join(current_path, "quant_cuda/quant.cu"),
                 ]
             )
else:
    quant_cuda = quant_cpu

__all__ = ['fixed_point_quantize', 'block_quantize', 'float_quantize', "quantizer"]

def assert_wl_fl(wl, fl, stage=""):
    if wl == -1 and fl != -1:
        raise ValueError("fixed point {} wl {}, fl {}".format(stage, wl, fl))

def get_module(x):
    if x.is_cuda:
        quant_module = quant_cuda
    else:
        quant_module = quant_cpu
    return quant_module

def quantizer(forward_number=None, backward_number=None,
              forward_rounding="stochastic", backward_rounding="stochastic",
              clamping_grad_zero=False, backward_hooks=[]):
    """
    Creates a quantization function to support quantizing forward and backward process differently.

    Args:
        - :param: forward_number (qtorch.Number, optional) : the number format used for forward quantization.
                  if is None, the quantization would be a identity mapping.
        - :param: backward_number (qtorch.Number, optional) : the number format used for backward quantization.
                  if is None, the quantization would be a identity mapping.
        - :param: forward_rounding (string) : rounding mode, \"stochastic\" or \"nearest\" (default: \"stochastic\")
        - :param: backward_rounding (string) : rounding mode, \"stochastic\" or \"nearest\" (default: \"stochastic\")
        - :param: clamping_grad_zero (bool) : zero out the gradient of numbers that are being clamped during forward propagation.
                  currently requires forward_number to be a fixed point number.
        - :param: backward_hooks (iterable) : iterable of functions that will be applied to gradients before backward quantization.
                  For example, this can be used to support custom scaling.

    Returns:
        A quantization function as specified (torch.Tensor -> torch.Tensor)
    """

    for rounding in [forward_rounding, backward_rounding]:
        assert rounding in ["stochastic", "nearest"], "invalid rounding type {:s}".format(rounding)
    for num in [forward_number, backward_number]:
        if num != None: assert isinstance(num, Number)

    if clamping_grad_zero==False:
        if forward_rounding=="nearest":
            if type(forward_number)==BlockFloatingPoint:
                forward_quant = lambda x, quant_module: quant_module.block_quantize_nearest(x, forward_number.wl, forward_number.dim)
            elif type(forward_number)==FixedPoint:
                forward_quant = lambda x, quant_module: quant_module.fixed_point_quantize_nearest(x, forward_number.wl,
                                                                                                  forward_number.fl, forward_number.clamp,
                                                                                                  forward_number.symmetric)
            elif type(forward_number)==FloatingPoint:
                forward_quant = lambda x, quant_module: quant_module.float_quantize_nearest(x, forward_number.man, forward_number.exp)
        elif forward_rounding=="stochastic":
            if type(forward_number)==BlockFloatingPoint:
                forward_quant = lambda x, quant_module: quant_module.block_quantize_stochastic(x, forward_number.wl, forward_number.dim)
            elif type(forward_number)==FixedPoint:
                forward_quant = lambda x, quant_module: quant_module.fixed_point_quantize_stochastic(x, forward_number.wl, forward_number.fl,
                                                                                                     forward_number.clamp, forward_number.symmetric)
            elif type(forward_number)==FloatingPoint:
                forward_quant = lambda x, quant_module: quant_module.float_quantize_stochastic(x, forward_number.man, forward_number.exp)
    else:
        if type(forward_number)==FixedPoint or forward_number==None:
            assert forward_number==None or forward_number.clamp == True, "must use clamping if zeroing out clamped gradient"
            if forward_rounding=="nearest":
                forward_quant = lambda x, quant_module: quant_module.fixed_point_quantize_nearest_mask(x, forward_number.wl, forward_number.fl, forward_number.symmetric)
            elif forward_rounding=="stochastic":
                forward_quant = lambda x, quant_module: quant_module.fixed_point_quantize_stochastic_mask(x, forward_number.wl, forward_number.fl, forward_number.symmetric)
        else:
            raise ValueError("zeroing clamping gradient only support fixed point.")

    if backward_rounding=="nearest":
        if type(backward_number)==BlockFloatingPoint:
            backward_quant = lambda a, quant_module: quant_module.block_quantize_nearest(a, backward_number.wl, backward_number.dim)
        elif type(backward_number)==FixedPoint:
            backward_quant = lambda a, quant_module: quant_module.fixed_point_quantize_nearest(a, backward_number.wl, backward_number.fl,
                                                                                               backward_number.clamp, backward_number.symmetric)
        elif type(backward_number)==FloatingPoint:
            backward_quant = lambda a, quant_module: quant_module.float_quantize_nearest(a, backward_number.man, backward_number.exp)
    elif backward_rounding=="stochastic":
        if type(backward_number)==BlockFloatingPoint:
            backward_quant = lambda a, quant_module: quant_module.block_quantize_stochastic(a, backward_number.wl, backward_number.dim)
        elif type(backward_number)==FixedPoint:
            backward_quant = lambda a, quant_module: quant_module.fixed_point_quantize_stochastic(a, backward_number.wl, backward_number.fl,
                                                                                                  backward_number.clamp, backward_number.symmetric)
        elif type(backward_number)==FloatingPoint:
            backward_quant = lambda a, quant_module: quant_module.float_quantize_stochastic(a, backward_number.man, backward_number.exp)

    if clamping_grad_zero == False:
        class Rounding(torch.autograd.Function):
            @staticmethod
            def forward(self, x):
                if forward_number==None: return x

                quant_module = get_module(x)
                out = forward_quant(x.contiguous(), quant_module)

                return out

            @staticmethod
            def backward(self, grad_output):
                if self.needs_input_grad[0]:
                    if backward_number == None:
                        grad_input = grad_output
                    else:
                        quant_module = get_module(grad_output)
                        grad_input = backward_quant(grad_output.contiguous(), quant_module)
                else:
                    grad_input = None

                return grad_input
    else:
        class Rounding(torch.autograd.Function):
            @staticmethod
            def forward(self, x):
                if forward_number==None:
                    self.mask = torch.zeros_like(x).byte()
                    return x
                else:
                    quant_module = get_module(x)
                    out, mask = forward_quant(x.contiguous(), quant_module)
                    self.mask = mask

                return out

            @staticmethod
            def backward(self, grad_output):
                if self.needs_input_grad[0]:
                    if backward_number == None:
                        grad_input = grad_output
                    else:
                        quant_module = get_module(grad_output)
                        # grad_output = grad_output.contiguous().masked_fill_(self.mask, 0)
                        for f in backward_hooks:
                            grad_output = f(grad_output)
                        grad_input = backward_quant(grad_output.contiguous(), quant_module).masked_fill(self.mask, 0)
                else:
                    grad_input = None

                return grad_input

    return Rounding.apply


def fixed_point_quantize(x, wl, fl, clamp=True, symmetric=False, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Fixed Point

    Args:
        - :param: `x` (torch.Tensor) :  the single precision number to be quantized
        - :param: `wl` (int) : word length of the fixed point number being simulated
        - :param: `fl` (int) : fractional length of the fixed point number being simulated
        - :param: `clamp` (bool, optional) : clamp input numbers into representable range. if false,
                  the quantization will only simulate the effect on precision
        - :param: `symmetric` (bool, optional) : discard the minimum representable number to make the representable
                  range symmetric
        - :param: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\" (default: \"stochastic\")

    Returns:
        - a quantized low-precision block floating point number (torch.Tensor)
    """
    assert isinstance(x, torch.Tensor)
    assert rounding in ["stochastic", "nearest"]
    assert_wl_fl(wl, fl)
    quant_module = get_module(x)
    if rounding == "nearest":
        out = quant_module.fixed_point_quantize_nearest(x.contiguous(), wl, fl, clamp, symmetric)
    elif rounding == "stochastic":
        out = quant_module.fixed_point_quantize_stochastic(x.contiguous(), wl, fl, clamp, symmetric)
    return out

def block_quantize(x, wl, dim=-1, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Block Floating Point

    Args:
        - :param: `x` (torch.Tensor) :  the single precision number to be quantized
        - :param: `wl` (int) : word length of the block floating point number being simulated
        - :param: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\"

    Returns:
        - a quantized low-precision block floating point number (torch.Tensor)
    """
    assert isinstance(x, torch.Tensor), "x is not a single precision Floating Point Tensor"
    assert rounding in ["stochastic", "nearest"], "invalid rounding mode, {}".format(rounding)
    quant_module = get_module(x)
    if rounding=="nearest":
        out = quant_module.block_quantize_nearest(x.contiguous(), wl, dim)
    elif rounding=="stochastic":
        out = quant_module.block_quantize_stochastic(x.contiguous(), wl, dim)
    return out

def float_quantize(x, exp, man, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Floating Point

    Args:
        - :attr: `x` (torch.Tensor) : the single precision number(torch.Tensor) to be quantized
        - :attr: `exp` (int) : number of bits allocated for exponent
        - :attr: `man` (int) : number of bits allocated for mantissa, not counting the virtual bit
        - :attr: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\"

    Returns:
        - a quantized low-precision floating point number (torch.Tensor)
    """
    assert isinstance(x, torch.Tensor), "x is not a single precision Floating Point Tensor"
    assert rounding in ["stochastic", "nearest"], "invalid rounding mode, {}".format(rounding)
    quant_module = get_module(x)
    if rounding=="nearest":
        out = quant_module.float_quantize_nearest(x.contiguous(), man, exp)
    elif rounding=="stochastic":
        out = quant_module.float_quantize_stochastic(x.contiguous(), man, exp)
    return out



import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint

class TestStochastic(unittest.TestCase):
    """
    invariant: stochastic rounding is unbiased
    """
    def calc_expectation_error(self, a, quant, N):
        b = torch.zeros_like(a)
        for i in range(int(N)):
            b = b * i/(i+1.) + quant(a) / (i+1)
        error = ((a-b)**2).mean().cpu().item()
        return error

    def test_stochastic_fixed(self):
        for wl, fl in [(7,6)]:
            for d in ['cpu', 'cuda']:
                a = torch.linspace(-0.9, 0.9, steps=100, device=d)
                quant = lambda x : fixed_point_quantize(x, wl=wl, fl=fl, clamp=True, symmetric=False)
                error = self.calc_expectation_error(a, quant, 1e5)
                self.assertTrue(error<1e-6)
                number = FixedPoint(wl=wl, fl=fl, clamp=True, symmetric=False)
                quant = quantizer(forward_number=number, forward_rounding="stochastic")
                error = self.calc_expectation_error(a, quant, 1e5)
                self.assertTrue(error<1e-6)

    def test_stochastic_block(self):
        for d in ['cpu', 'cuda']:
            for dim in [-1, 0, 1]:
                a = torch.rand(10, 10, 10)
                quant = lambda x : block_quantize(x, wl=6, dim=dim)
                error = self.calc_expectation_error(a, quant, 1e5)
                self.assertTrue(error < 1e-4)
                number = BlockFloatingPoint(wl=6, dim=dim)
                quant = quantizer(forward_number=number, forward_rounding="stochastic")
                error = self.calc_expectation_error(a, quant, 1e5)
                self.assertTrue(error<1e-4)

    def test_stochastic_float(self):
        for d in ['cpu', 'cuda']:
            a = torch.rand(100).to(device=d)
            quant = lambda x : float_quantize(x, exp=6, man=5)
            error = self.calc_expectation_error(a, quant, 1e5)
            self.assertTrue((error < 1e-6))
            number = FloatingPoint(exp=6, man=5)
            quant = quantizer(forward_number=number, forward_rounding="stochastic")
            error = self.calc_expectation_error(a, quant, 1e5)
            self.assertTrue(error<1e-6)

if __name__ == "__main__":
    unittest.main()



import torch
from torch.optim import Optimizer, SGD, Adam

__all__ = ["OptimLP"]

class OptimLP(Optimizer):
    """
    A low-precision optimizer wrapper that handles weight, gradient, accumulator quantization.

    Args:
        - :attr: `optim`: underlying optimizer to use
        - :attr: `weight_quant`: a weight quantization function which takes a pytorch tensor and returns a tensor. If None, does not quantize weight.
        - :attr: `grad_quant`: a gradient quantization function which takes a pytorch tensor and returns a tensor. If None, does not quantize weight.
        - :attr: `grad_scaling`: float, scaling factor before apply gradient quantization.
        - :attr: `momentum_quant`: a momentum quantization function which takes a pytorch tensor and returns a tensor.
                                   If None, does not quantize weight.
        - :attr: `acc_quant`: a accumulator quantization function which takes
                              a pytorch tensor and returns a tensor. If not None, a
                              OptimLP object would create memory copies of model parameters that serve as
                              gradient accumulators. If None, does not use gradient accumulators.

    Example:
        >>> weight_q = quantizer(...) # define weight quantization
        >>> optimizer = SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer = OptimLP(optiimizer, weight_quant=weight_q)
    """

    def __init__(self, optim,
                 weight_quant=None,
                 grad_scaling=1.0,
                 grad_quant=None,
                 momentum_quant=None,
                 acc_quant=None):
        assert (isinstance(optim, SGD) or isinstance(optim, Adam))
        super(OptimLP, self).__init__(optim.param_groups, optim.defaults) # place holder

        # python dictionary does not copy by default
        self.param_groups = optim.param_groups
        self.optim = optim

        assert grad_scaling > 0, "gradient scaling must be positive"
        self.grad_scaling = grad_scaling

        self.weight_quant=weight_quant
        self.grad_quant=grad_quant
        self.momentum_quant=momentum_quant
        self.acc_quant=acc_quant

        if isinstance(self.optim, SGD):
            self.momentum_keys = ['momentum_buffer']
        elif isinstance(self.optim, Adam):
            # TODO: support amsgrad
            self.momentum_keys = ['exp_avg', 'exp_avg_sq']
        else:
            raise NotImplementedError("Only supporting Adam and SGD for now. ")

        if self.acc_quant != None:
            self.weight_acc = {}
            for group in self.param_groups:
                for p in group['params']:
                    self.weight_acc[p] = p.detach().clone()

    def step(self, closure=None):
        """
        Performs one step of optimization with the underlying optimizer.
        Quantizes gradient and momentum before stepping. Quantizes gradient accumulator and weight after stepping.
        """
        # quantize gradient
        if not self.grad_quant is None:
            for group in self.param_groups:
                for p in group['params']:
                    p.grad.data = self.grad_quant(p.grad.data*self.grad_scaling)

        # switch acc into weight before stepping
        if not self.acc_quant is None:
            for group in self.param_groups:
                for p in group['params']:
                    p.data = self.weight_acc[p].data

        loss = self.optim.step()

        # switch weight into acc after stepping and quantize
        if not self.acc_quant is None:
            for group in self.param_groups:
                for p in group['params']:
                    p.data = self.weight_acc[p].data = self.acc_quant(p.data).data

        # quantize weight from acc
        if not self.weight_quant is None:
            for group in self.param_groups:
                for p in group['params']:
                    p.data = self.weight_quant(p.data).data

        # quantize momentum
        if not self.momentum_quant is None:
            for group in self.param_groups:
                if group['momentum'] == 0: continue
                for p in group['params']:
                    param_state = self.optim.state[p]
                    for key in self.momentum_keys:
                        param_state[key] = self.momentum_quant(param_state[key])

        return loss

    def __repr__(self):
        return "LP Optimizer: {}".format(self.optim.__repr__())

    def __str__(self):
        return "LP Optimizer: {}".format(self.optim.__str__())
